/*                                                             */
/*   Calcul de la valeur de PI par approximation numerique :   */
/*   / 1     1                     1               --          */
/*   |   ----------  dx = arctan(x)  =  arctan(1)= ||/4        */
/*   |    1 + x**2                 0                           */
/*   / 0                                                       */
/*                       par la methode des trapezes.          */
/*                                                             */
/*      Auteur   : Isabelle DUPAYS (CNRS/IDRIS - France)       */     
/*                        <Isabelle.Dupays@idris.fr>           */ 
/*      Cree le  : Thu Oct  2 09:43:29 MET 1997                */

#include "mpi.h"
#include <math.h>
#include <stdio.h>

#define M_PI        3.14159265358979323846
#define PI_div_1000 0.00314159265358979323
#define NbTERM      5000

/* Fonction a integrer */
double f(double a)
{
  return (4.0 / (1.0 + a*a));
}

int main(int argc,char *argv[])
{
  int nb_procs,rang, K_procs;
  int n;
  int i;
  double mon_pi, pi, h, somme, x;
  double t1,t2;
  int KR, KR_bcast, KR_reduce;
  int etiquette = 100;
  
  float TabF[ NbTERM ];
  float TmpF[ NbTERM ];
  float VLR;

  char procname[ 1024 ]; 
  char hostname[ 1024 ]; 
  unsigned long tailleBuf = 1014;
  int resultlen;

  /* Initialisation de MPI */
  MPI_Init(&argc,&argv);

  /* Connaitre le nombre total de processeurs  */
  MPI_Comm_size(MPI_COMM_WORLD,&nb_procs);

  /* Savoir quel processeur je suis */
  MPI_Comm_rank(MPI_COMM_WORLD,&rang);

  KR = MPI_Get_processor_name( procname, &resultlen );
  printf("%s> Processus Rank : %d/%d\n", procname, rang, nb_procs);
  
  if ( rang == 0 )  {
    n = 100; // valeur emise en broadcast      

    int   KC; 
    float signe, Uk, Sk;
    signe = -1.0;

	Uk = 1.0; Sk = 0.0;
  
	for ( KC=0; KC < NbTERM; KC++) {
      Uk = signe * (float) cos( KC * PI_div_1000  ) * KC;
      Sk = Sk + Uk;
	  TabF[ KC ] = Sk;
	  //TabF[ KC ] = signe * KC;
	  signe = - signe;
    }
  }

  if ( rang == 0 )  {
	for ( K_procs = 1; K_procs < nb_procs; K_procs++ ) {
	  MPI_Status status;
	  MPI_Recv( &hostname, sizeof( hostname ), MPI_CHAR, MPI_ANY_SOURCE, etiquette, MPI_COMM_WORLD, &status );
	  printf("%2.2d : %s\n", status.MPI_SOURCE , hostname);
	}
    t1 = MPI_Wtime(); //-- Temps initial en seconde
  } else {
	 MPI_Request Request;
     MPI_Isend( &procname, sizeof( hostname ), MPI_CHAR, 0, etiquette, MPI_COMM_WORLD, &Request );
     //MPI_Ssend( &rang, 1 ,MPI_INTEGER, 0, etiquette, MPI_COMM_WORLD );
  }
  etiquette++;
  
  int NbDATA = NbTERM / nb_procs;
  
  //-- Envoi du tableau entier  tous les processus
  KR_bcast = MPI_Bcast(&TabF, NbDATA, MPI_FLOAT, 0, MPI_COMM_WORLD);
  //MPI_Scatter( TabF, NbDATA, MPI_FLOAT, TmpF, NbDATA, MPI_INT, 0, MPI_COMM_WORLD);

  //-- Verification
  if ( rang == 0 )  {
	for ( K_procs = 1; K_procs < nb_procs; K_procs++ ) {
	  MPI_Status status;
	  MPI_Recv( &VLR, 1, MPI_FLOAT, MPI_ANY_SOURCE, etiquette, MPI_COMM_WORLD, &status );
	  printf("Verif> %2.2d : %15.8f\n", status.MPI_SOURCE , VLR );
	}
  } else {
     VLR = TabF[ rang ];
	 MPI_Request Request;
     MPI_Isend( &VLR, 1, MPI_FLOAT, 0, etiquette, MPI_COMM_WORLD, &Request );
  }
  etiquette++;

  int K_Cur;
  int K_Vlr = rang;
  int K_Deb = rang + nb_procs;
  int K_Pas = (NbTERM / nb_procs) -1;
  int K_Fin = K_Deb + nb_procs*K_Pas;

  while ( K_Fin < NbTERM) { 
	K_Fin += nb_procs;
  }
    //-- Verification K_Deb
  if ( rang == 0 )  {
	for ( K_procs = 1; K_procs < nb_procs; K_procs++ ) {
	  MPI_Status status;
	  MPI_Recv( &K_Cur, 1, MPI_INT, MPI_ANY_SOURCE, etiquette, MPI_COMM_WORLD, &status );
	  printf("K_Deb> %2.2d : %3.3d\n", status.MPI_SOURCE , K_Cur );
	}
  } else {
	 MPI_Request Request;
     MPI_Isend( &K_Deb, 1, MPI_INT, 0, etiquette, MPI_COMM_WORLD, &Request );
  }
  etiquette++;

  //-- Verification K_Fin
  if ( rang == 0 )  {
	for ( K_procs = 1; K_procs < nb_procs; K_procs++ ) {
	  MPI_Status status;
	  MPI_Recv( &K_Cur, 1, MPI_INT, MPI_ANY_SOURCE, etiquette, MPI_COMM_WORLD, &status );
	  printf("K_Fin> %2.2d : %3.3d\n", status.MPI_SOURCE , K_Cur );
	}
  } else {
	 MPI_Request Request;
     MPI_Isend( &K_Fin, 1, MPI_INT, 0, etiquette, MPI_COMM_WORLD, &Request );
  }
  etiquette++;

  MPI_Barrier( MPI_COMM_WORLD );

  //-- TRI Shell partiel

  for ( K_Cur = K_Deb; K_Cur < K_Fin; K_Cur=K_Cur+nb_procs ) {
    K_Vlr = K_Cur - nb_procs;
    VLR= TabF[ K_Vlr ];
  
    if ( TabF[ K_Cur ] > VLR ) {
      TabF[ K_Vlr ] = TabF[ K_Cur ];
      TabF[ K_Cur ] = VLR;
    }
  } 

  //-- Verification TRI
  if ( rang == 0 )  {


  }

  //-- Calcul de PI : Envoi de la valeur de n aux autres processeurs  */   
  
  KR_bcast = MPI_Bcast(&n, 1, MPI_INT, 0, MPI_COMM_WORLD);
  if ( KR_bcast != 0 ) {
	printf("MPI_BCAST> PE:%d -> %d\n", rang, KR_bcast );
  }

  /* Calcul de la longueur de l'intervalle  */
  h   = 1.0 / (double) n;
  somme = 0.0;

  /* Chaque processus calcule une somme partielle */
  for (i = rang + 1; i <= n; i += nb_procs) {
    x = h * ((double)i - 0.5);
    somme += f(x);
  }
  mon_pi = h * somme;

  /* Calcul de la somme globale par le processeur 0 */
  KR_reduce = MPI_Reduce(&mon_pi, &pi, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if ( KR_reduce != 0 ) {
	printf("MPI_REDUCE> PE:%d -> %d\n", rang, KR_reduce );
  }

  if (rang == 0) {
    /* Affichage du resultat par le processeur 0 */
    printf("pi vaut environ %.16f, erreur  %.16f\n",
	   pi, fabs(pi - M_PI));
    /* Mesure du temps en seconde   */      
    t2 = MPI_Wtime();
    printf("Temps total = %f\n",t2 - t1);
  }

  /* Desactivation de MPI */
  MPI_Finalize();

  return(0);

}

