!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! -*- Mode: F90 -*- !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! CommCartSub.f90 --- Subdiviser une grille 2D avec MPI_CART_SUB()
!! 
!! Auteur          : Jalel Chergui (CNRS/IDRIS - France) <Jalel.Chergui@idris.fr>
!! Cree le         : Tue Apr  2 16:29:22 1996
!! Dern. mod. par  : Jalel Chergui (CNRS/IDRIS - France) <Jalel.Chergui@idris.fr>
!! Dern. mod. le   : Wed Jan 15 19:39:13 1997
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

program CommCartSub
  implicit none
  include 'mpif.h'
  integer                        :: CommCart2D,CommCart1D
  integer,parameter              :: NDimCart2D=2
  integer, dimension(NDimCart2D) :: DimCart2D,CoordCart2D
  logical, dimension(NDimCart2D) :: Periode, DimReste
  logical                        :: Reordonne
  integer                        :: nb_procs,rang,code
  integer,parameter              :: m=4
  real, dimension(m)             :: V
  real                           :: W
  
  call MPI_INIT(code)
  call MPI_COMM_SIZE(MPI_COMM_WORLD, nb_procs, code)
  
  if ( nb_procs .NE. 12 ) then
    print*, 'nombre de processeurs pour cette execution doit etre 12 !'
    call MPI_FINALIZE(code)
    stop
  endif
    
  !*** Creation de la grille 2D initiale
  DimCart2D(1) = 3
  DimCart2D(2) = 4
  if(DimCart2D(1)*DimCart2D(2) /= nb_procs) stop
  Periode(:)   = .false.
  ReOrdonne    = .false.
  call MPI_CART_CREATE(MPI_COMM_WORLD, NDimCart2D, DimCart2D, &
                       Periode, ReOrdonne, CommCart2D, code)
  call MPI_COMM_RANK(CommCart2D, rang, code)
  call MPI_CART_COORDS(CommCart2D, rang, NDimCart2D,  &
                       CoordCart2D, code)
  
  !*** Initialisation du vecteur V et du scalaire W
  V(:) = 0. ; W = 0.
  if (CoordCart2D(2) == 1) V(:)=real(rang)
  
  !*** Nous voulons que Chaque ligne de la grille soit
  !*** une topologie cartesienne 1D.
  DimReste(1) = .false. ; DimReste(2)= .true.
  
  !*** Subdivision de la grille cartesienne 2D.
  call MPI_CART_SUB(CommCart2D,DimReste,CommCart1D,code)
  
  !*** Les processus de la colonne 2 diffusent selectivement
  !*** le vecteur V aux processus de leur ligne. 
  call MPI_SCATTER(V,1,MPI_REAL,W,1,MPI_REAL,1,CommCart1D,code)
  
  print '("Rang : ",I2," ; Coordonnees : (",I1,",",I1,") ; W = ",F2.0)', &
         rang,CoordCart2D(1),CoordCart2D(2),W
    
  !*** Destruction des communicateurs
  call MPI_COMM_FREE(CommCart1D,code)
  call MPI_COMM_FREE(CommCart2D,code)
  
  call MPI_FINALIZE(code)
end program CommCartSub
